/*
** Module   : DefaultDirtyShareManager.java
** Abstract : Default implementation of DirtyShareManager
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20080327   @38627 Created initial version.  Default
**                           implementation of DirtyShareManager.
** 002 ECF 20080606   @38639 Fixed NPE. Added safety checks in cleanup
**                           code.
** 003 ECF 20080612   @38716 Replaced isEntityDirty() with
**                           getEntityDirtyState(). New method returns
**                           state information (is entity dirty with
**                           changes and/or deletes), rather than a simple
**                           boolean.
** 004 CA  20080626   @39072 When rollbackInsert is called, remove the ID
**                           from the changes map.
** 005 CA  20080715   @39109 Fixed removeRecords - the session must be 
**                           closed before the method is finished.
** 006 CA  20080716   @39121 Renamed method lockUniqueIndexes to 
**                           lockAllIndexes - this will lock all the 
**                           indexes for a certain entity. Modified the
**                           delete(...) and rollbackInsert(...) methods
**                           so the index locking can be disabled.
** 007 ECF 20080815   @39543 Added support for global change events. These
**                           are events that represent index-changing
**                           inserts, updates, or deletes, which are
**                           processed by sessions other than the session
**                           in which the event originated.
** 008 CA  20080815   @39445 Support API change in DBUtils.
** 009 ECF 20090428   @42130 Store a copy of all inserted DMOs. This ensures a
**                           DMO from a lookup is in the same state it was
**                           when first inserted. Modified getDirtyDMO()
**                           accordingly.
** 010 ECF 20090512   @42157 Added assertion to getDirtyInfo().
** 011 ECF 20090513   @42169 Modified variant of update() method. Added lock
**                           parameter to make index locking optional.
** 012 ECF 20090601   @42571 Added support for Progress isolation leak quirk.
**                           Added new methods. Changed implementations of
**                           getDirtyDMO() and getDirtyInfo().
** 013 ECF 20090609   @42640 Modified cleanupChanges(). The list of entities
**                           which have had their uncommitted changes leaked
**                           is now passed in, allowing proper cleanup.
** 014 ECF 20090611   @42662 Optimized getDirtyDMO(). Avoid overhead of index
**                           locking if requested DMO is not being tracked.
** 015 ECF 20090701   @43226 Removed faulty assert statement. Added debug
**                           logging.
** 016 ECF 20090720   @43308 Fixed delete(). If the deleted record already is
**                           being tracked, it must be removed from the dirty
**                           database.
** 017 ECF 20090723   @43340 Fixed methods which execute queries. Parameters
**                           to these queries require preprocessing before
**                           being bound.
** 018 ECF 20100317   @44752 Added protection logic to removeRecords to avoid
**                           IllegalArgumentException. This is a workaround
**                           for a situation that causes a null DMO to come
**                           back from a call to session.get() for a primary
**                           key that should be valid. Root cause currently is
**                           unknown, but this change should protect against
**                           abend at this location.
** 019 SVL 20130331          Upgraded to Hibernate 4.
** 020 ECF 20131001          Added metadata support.
** 021 ECF 20131028          Import change.
** 022 SVL 20140210          Set JPA-style HQL parameters instead of positional.
** 023 ECF 20140825          Track unvalidated inserts. Replaced Apache commons logging with J2SE
**                           logging.
*/

package com.goldencode.p2j.persist.dirty;

import java.io.*;
import java.lang.reflect.*;
import java.sql.*;
import java.util.*;
import java.util.logging.*;
import org.hibernate.*;
import org.hibernate.dialect.*;
import org.hibernate.internal.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.util.*;

/**
 * Default implementation of {@link DirtyShareManager}, to be used for local,
 * primary databases (database instances for which the current P2J server is
 * authoritative).  Stores and shares knowledge of uncommitted modifications
 * to database records, specifically, those modifications which affect
 * database indexes.
 * <p>
 * Information about uncommitted inserts and updates to existing records is
 * managed at two levels.  A modified copy of the inserted or updated record
 * is stored in a backing, "dirty" database.  In addition, a set of in-memory
 * data structures maintains information about the tables (or more accurately,
 * the DMO entity names) and primary keys of affected records.  The former
 * allows the detailed, uncommitted index updates of a record to be queried
 * and retrieved if necessary.  However, the latter enables a quick, high
 * level screening capability to avoid the need for the slower database access
 * in many common cases.
 * <p>
 * Information about uncommitted deletes is managed only in an in-memory data
 * structure, because it is unnecessary to store the full record in the case
 * of an uncommitted delete in one context;  all other contexts still have
 * access to the full record.  It is simply the knowledge of the delete action
 * that must be managed and shared.
 * <p>
 * Concurrent access to the uncommitted delete information and to the high
 * level screening structures for the insert and update information is
 * managed via simple monitors by using synchronized blocks on 3 lock objects,
 * one for each action type.
 * <p>
 * To maximize concurrent access when it is determined that the dirty database
 * must be consulted, this implementation uses a lock manager to serialize
 * read (share) and write (exclusive) access to the backing database.
 * Granularity of access for record inserts and updates is determined by the
 * database indexes in use for a particular update or query.  For instance, if
 * a query is walking a particular index for its results, only a share lock on
 * that single index is acquired when checking for an uncommitted record which
 * might override a query's primary result.  However, when a record is
 * updated, an exclusive lock must be acquired on every index affected by all
 * of the updated DMO properties, before the dirty record can be updated in
 * the dirty database.  For a newly inserted record, every index on the
 * associated table must be exclusively locked for the uncommitted insert to
 * be recorded in the dirty database.
 * 
 * @see  DefaultDirtyShareContext
 */
public class DefaultDirtyShareManager
implements DirtyShareManager
{
   /** Bogus record ID for locking at index level */
   private static final int INDEX_ID = 0;
   
   /** Name of primary key property */
   private static final String PK_PROPERTY = DatabaseManager.PRIMARY_KEY;
   
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(DirtyShareManager.class);
   
   /** Associated dirty database */
   private final Database dirtyDatabase;
   
   /** Dialect of the dirty database */
   private final P2JDialect dialect;
   
   /** Hibernate session factory */
   private final SessionFactory sessionFactory;
   
   /** Helper object to lookup indexes */
   private final IndexHelper indexHelper;
   
   /** Set of DMO entity names for tables which are already created */
   private final Set<String> createdTables = new HashSet<String>();
   
   /** Index lock manager */
   private final InMemoryLockManager lockManager;
   
   /** Index event manager */
   private final GlobalEventManager eventManager;
   
   /** Lock for deletes map */
   private final Object deleteLock = new Object();
   
   /** Lock for inserts map */
   private final Object insertLock = new Object();
   
   /** Lock for changes map */
   private final Object changeLock = new Object();
   
   /** Map of record identifiers to inserted records */
   private Map<RecordIdentifier, Persistable> inserts = new HashMap<>();
   
   /** Map of record identifiers for new inserts to set of indexes requiring validation */
   private Map<RecordIdentifier, Set<String>> unvalidatedInserts = new HashMap<>();
   
   /** Dirty update ids by entity name */
   private Map<String, Set<Serializable>> changes = new HashMap<>();
   
   /** Dirty deleted ids by entity name */
   private Map<String, Set<Serializable>> deletes = new HashMap<>();
   
   /**
    * Constructor.  Looks up the dirty database associated with the given
    * primary database, and retrieves its <code>SessionFactory</code>.
    * Retrieves an <code>IndexHelper</code> instance and creates a headless
    * <code>LockManager</code> to coordinate read and write access to the
    * dirty database.
    * 
    * @param   database
    *          Primary database for which we are managing a dirty database.
    */
   DefaultDirtyShareManager(Database database)
   {
      Database dirty = new Database(database, Database.Type.DIRTY);
      Dialect dialect = DatabaseManager.getDialect(dirty);
      if (dialect == null)
      {
         throw new IllegalArgumentException(
            "Cannot determine dialect for dirty database " + dirty);
      }
      try
      {
         this.dialect = (P2JDialect) DatabaseManager.getDialect(dirty);
      }
      catch (ClassCastException exc)
      {
         throw new IllegalArgumentException(
            "Unsupported dirty database dialect:  " + dialect);
      }
      
      this.dirtyDatabase = dirty;
      this.sessionFactory = DatabaseManager.getSessionFactory(dirty);
      this.indexHelper = IndexHelper.get(database);
      this.eventManager = new GlobalEventManager(database);
      this.lockManager = new InMemoryLockManager()
      {
         protected boolean isHeadless() { return true; }
      };
      lockManager.setDatabase(dirty);
   }
   
   /**
    * Register with a global event clearinghouse an interest in insert,
    * delete, and update events which affect the given DMO entity.
    * 
    * @param   entity
    *          DMO entity for which the caller is interested in tracking
    *          change.
    * 
    * @return  A unique identifier for the registrant, assigned by the event
    *          clearinghouse, used later to collect events and deregister.
    * 
    * @see     #getGlobalEvents(long, long)
    * @see     #deregisterForGlobalEvents(long)
    */
   public long registerForGlobalEvents(String entity)
   {
      return eventManager.register(entity);
   }
   
   /**
    * Deregister with a global event clearinghouse an interest in insert,
    * delete, and update events which affect a particular DMO type.
    * 
    * @param   registerID
    *          A unique identifier for the registrant being deregistered.
    * 
    * @see     #getGlobalEvents(long, long)
    * @see     #registerForGlobalEvents(String)
    */
   public void deregisterForGlobalEvents(long registerID)
   {
      eventManager.deregister(registerID);
   }
   
   /**
    * Retrieve any new events that have been collected by a global event
    * clearinghouse, representing inserts, deletes, and updates made to
    * record types for which the caller previously has registered interest.
    * 
    * @param   registerID
    *          A unique identifier for an object which has registered interest
    *          in record-changing events.
    * @param   lastID
    *          A unique identifier representing the last event retrieved by
    *          the caller.  If less than 0, it is assumed no events have yet
    *          been retrieved since the caller registered interest in these
    *          events.
    * 
    * @return  Array of applicable, new events, or <code>null</code> if no
    *          such events have been collected.
    * 
    * @throws  EventRegistrationException
    *          if <code>registerID</code> is invalid or has expired.
    */
   public GlobalChangeEvent[] getGlobalEvents(long registerID, long lastID)
   throws EventRegistrationException
   {
      return eventManager.getEvents(registerID, lastID);
   }
   
   /**
    * Lock or unlock all indexes associated with the given DMO entity.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   lockType
    *          Type of lock to acquire, or <code>NONE</code> to release
    *          existing locks.  <code>NO_WAIT</code> variants will be treated
    *          as regular, blocking variants.
    * 
    * @throws  PersistenceException
    *          if there is an error querying JDBC metadata while retrieving
    *          index information.
    */
   public void lockAllIndexes(String entity, LockType lockType)
   throws PersistenceException
   {
      List<String> indexes = indexHelper.getIndexesForEntity(entity, false);
      lockMultiple(indexes, lockType);
   }
   
   /**
    * Begin tracking a newly inserted record in the dirty database.
    * <p>
    * This record is removed when the adding context's current transaction is
    * committed or rolled back.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation class associated
    *          with a table being tracked for uncommitted changes.
    * @param   copy
    *          Copy of the record to be tracked, ready to be inserted into the dirty database.
    * @param   uniqueIndexes
    *          If the insertion is happening at record creation, and the record has not yet been
    *          validated, this will represent the collection of unique indexes that will need to
    *          be validated before the record can safely be persisted. If <code>null</code>, this
    *          indicates the record has been fully validated and should no longer be tracked as
    *          unvalidated.
    * @param   lock
    *          <code>true</code> to acquire write locks to insert the record; <code>false</code>
    *          to assume write locks are acquired by calling code.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void insert(String entity,
                      Persistable copy,
                      Collection<String> uniqueIndexes,
                      boolean lock)
   throws PersistenceException
   {
      List<String> lockIndexes = (lock ? indexHelper.getIndexesForEntity(entity, false) : null);
      boolean early = (uniqueIndexes != null);
      trackChange(entity, copy, lockIndexes, null, true, early, uniqueIndexes);
   }
   
   /**
    * Notify the dirty share manager that the given record has been validated, and should no
    * longer be tracked as unvalidated.
    * 
    * @param   ident
    *          Record identifier (using entity name and primary key).
    * @param   copy
    *          Updated copy of DMO for the dirty manager to share with other contexts.
    */
   public void notifyInsertValidated(RecordIdentifier ident, Persistable copy)
   {
      synchronized (insertLock)
      {
         unvalidatedInserts.remove(ident);
         inserts.put(ident, copy);
      }
   }
   
   /**
    * Indicate whether the record identified by <code>ident</code> represents a newly inserted
    * record, which has an unvalidated index, <code>index</code>. Assumes appropriate indexes
    * already are locked by the caller.
    * 
    * @param   ident
    *          Record identifier (using entity name and primary key).
    * @param   index
    *          Index to be checked for validation.
    * 
    * @return  <code>true</code> if <code>ident</code> represents a new record in a foreign
    *          context with unvalidated index <code>index</code>.
    */
   public boolean isUnvalidatedIndex(RecordIdentifier ident, String index)
   {
      synchronized (insertLock)
      {
         Set<String> unchecked = unvalidatedInserts.get(ident);
         
         return (unchecked != null && unchecked.contains(index));
      }
   }
   
   /**
    * Rollback the insert of a record previously introduced with the {@link
    * #insert(String, Persistable, boolean, boolean)} method.
    * 
    * @param   ident
    *          Identifier for the target record, which encapsulates its entity
    *          name and primary key.
    * @param   lock
    *          <code>true</code> to acquire write locks;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void rollbackInsert(RecordIdentifier ident, boolean lock)
   throws PersistenceException
   {
      String entity = ident.getTable();
      Serializable id = ident.getRecordID();
      Session session = null;
      Transaction tx = null;
      List<String> indexes = (lock ? indexHelper.getIndexesForEntity(entity, false) : null);
      
      try
      {
         if (indexes != null)
         {
            lockMultiple(indexes, LockType.EXCLUSIVE);
         }
         
         synchronized (insertLock)
         {
            inserts.remove(ident);
            unvalidatedInserts.remove(ident);
         }
         
         synchronized (changeLock)
         {
            Set<Serializable> changed = changes.get(entity);
            changed.remove(id);
            
            if (changed.isEmpty())
            {
               changes.remove(entity);
            }
         }
         
         session = sessionFactory.openSession();
         tx = session.beginTransaction();
         Persistable copy = (Persistable) session.get(entity, id);
         session.delete(copy);
         tx.commit();
      }
      catch (HibernateException exc)
      {
         if (tx != null)
         {
            tx.rollback();
         }
         
         DBUtils.handleException(dirtyDatabase, exc);
         
         throw new PersistenceException(exc);
      }
      finally
      {
         if (indexes != null)
         {
            lockMultiple(indexes, LockType.NONE);
         }
         
         if (session != null)
         {
            try
            {
               session.close();
            }
            catch (Exception exc)
            {
               DBUtils.handleException(dirtyDatabase, exc);
            }
         }
         
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE,
                    "Rolled back insert of " + ident + "; currently tracking " + changes);
         }
      }
   }
   
   /**
    * Indicate whether the record specified by the given entity and primary
    * key is being tracked for changes within an uncommitted transaction.
    * This includes records which have been added to the dirty database due to
    * an insert or an update, not due to a delete.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the record to be checked.
    * 
    * @return  <code>true</code> to indicate the target record is being
    *          tracked for changes, else <code>false</code>.
    */
   public boolean isTracked(String entity, Serializable id)
   {
      synchronized (changeLock)
      {
         Set<Serializable> changed = changes.get(entity);
         
         return (changed != null && changed.contains(id));
      }
   }
   
   /**
    * Begin tracking a DMO for changes, based on the updates indicated by the
    * <code>properties</code> and <code>values</code> arguments.
    * <p>
    * Make a copy of the given record, apply the specified updates to that
    * copy, and store the copy in the dirty database.
    * <p>
    * This record is removed when the adding context's current transaction is
    * committed or rolled back.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   copy
    *          Copy of the record to be tracked.
    * @param   properties
    *          Array of names of those DMO properties which were updated.
    *          These should represent columns which participate in one or more
    *          indexes.
    * @param   values
    *          Updated DMO property values.  Each value in the array
    *          corresponds positionally with its matching property name in the
    *          <code>properties</code> array.
    * @param   validatedIndexes
    *          Collection of unique indexes which have been checked successfully for this update
    *          to have passed validation.
    * @param   lock
    *          <code>true</code> to acquire write locks to update the record;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void update(String entity,
                      Persistable copy,
                      String[] properties,
                      BaseDataType[] values,
                      Collection<String> validatedIndexes,
                      boolean lock)
   throws PersistenceException
   {
      applyUpdates(entity, copy, properties, values);
      
      List<String> indexes = (lock ? indexesForProperties(entity, properties) : null);
      trackChange(entity, copy, indexes, properties, false, false, validatedIndexes);
   }
   
   /**
    * Retrieve a DMO which is being tracked in the dirty database, and apply
    * the specified updates to it, as provided by the <code>properties</code>
    * and <code>values</code> arguments.  It is assumed the DMO already exists
    * in the database, from a previous call to {@link
    * #update(String, Persistable, String[], BaseDataType[], boolean)}.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key id of the record to be retrieved and updated.
    * @param   properties
    *          Array of names of those DMO properties which were updated.
    *          These should represent columns which participate in one or more
    *          indexes.
    * @param   values
    *          Updated DMO property values.  Each value in the array
    *          corresponds positionally with its matching property name in the
    *          <code>properties</code> array.
    * @param   validatedIndexes
    *          Collection of unique indexes which have been checked successfully for this update
    *          to have passed validation. May be <code>null</code>.
    * @param   lock
    *          <code>true</code> to acquire write locks to update the record;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void update(String entity,
                      Serializable id,
                      String[] properties,
                      BaseDataType[] values,
                      Collection<String> validatedIndexes,
                      boolean lock)
   throws PersistenceException
   {
      List<String> indexes = (lock ? indexesForProperties(entity, properties) : null);
      Session session = null;
      Transaction tx = null;
      
      try
      {
         if (lock)
         {
            lockMultiple(indexes, LockType.EXCLUSIVE);
         }
         
         session = sessionFactory.openSession();
         tx = session.beginTransaction();
         Persistable copy = (Persistable) session.get(entity, id);
         applyUpdates(entity, copy, properties, values);
         tx.commit();
         
         eventManager.enqueueUpdate(entity, copy, properties);
         
         synchronized (changeLock)
         {
            Set<Serializable> changed = changes.get(entity);
            if (changed == null)
            {
               changed = new HashSet<Serializable>();
               changes.put(entity.intern(), changed);
            }
            changed.add(id);
         }
         
         if (validatedIndexes != null)
         {
            synchronized (insertLock)
            {
               RecordIdentifier ident = new RecordIdentifier(entity, id);
               Set<String> unchecked = unvalidatedInserts.get(ident);
               if (unchecked != null)
               {
                  if (unchecked.removeAll(validatedIndexes) && unchecked.isEmpty())
                  {
                     unvalidatedInserts.remove(ident);
                  }
               }
            }
         }
      }
      catch (HibernateException exc)
      {
         if (tx != null)
         {
            tx.rollback();
         }
         
         DBUtils.handleException(dirtyDatabase, exc);
         
         throw new PersistenceException(exc);
      }
      finally
      {
         if (lock)
         {
            lockMultiple(indexes, LockType.NONE);
         }
         
         if (session != null)
         {
            try
            {
               session.close();
            }
            catch (Exception exc)
            {
               DBUtils.handleException(dirtyDatabase, exc);
            }
         }
      }
   }
   
   /**
    * Overwrite the state of the records in the dirty database which match the
    * IDs of the given list of DMOs for the given entity, and store a snapshot
    * of each DMO in the map of newly inserted records.
    * <p>
    * This is necessary to support a quirk/bug in Progress which compromises
    * transaction isolation by leaking uncommitted changes (even non-indexed
    * data) under certain conditions.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   dmos
    *          Array of DMOs which are to be merged into their counterpart
    *          records in the dirty database and whose snapshots are to be
    *          stored in the map of newly inserted records.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void updateSnapshots(String entity, Persistable[] dmos)
   throws PersistenceException
   {
      List<String> indexes = indexHelper.getIndexesForEntity(entity, false);
      Session session = null;
      Transaction tx = null;
      
      try
      {
         lockMultiple(indexes, LockType.EXCLUSIVE);
         
         session = sessionFactory.openSession();
         tx = session.beginTransaction();
         
         for (Persistable dmo : dmos)
         {
            session.merge(entity, dmo);
         }
         tx.commit();
         
         synchronized (insertLock)
         {
            for (Persistable dmo : dmos)
            {
               // We force the inserts map to do double duty here.  Even
               // though these snapshots do not necessarily represent newly
               // inserted records, putting them in this map makes all
               // downstream processing behave properly, since we want these
               // snapshots returned rather than the actual records stored in
               // the database, when the DirtyShareContext requests a dirty
               // check.
               Persistable copy = (Persistable) dmo.deepCopy();
               Serializable id = dmo.getId();
               RecordIdentifier ident = new RecordIdentifier(entity, id);
               inserts.put(ident, copy);
            }
         }
      }
      catch (HibernateException exc)
      {
         if (tx != null)
         {
            tx.rollback();
         }
         
         DBUtils.handleException(dirtyDatabase, exc);
         
         throw new PersistenceException(exc);
      }
      finally
      {
         lockMultiple(indexes, LockType.NONE);
         
         if (session != null)
         {
            try
            {
               session.close();
            }
            catch (Exception exc)
            {
               DBUtils.handleException(dirtyDatabase, exc);
            }
         }
      }
   }
   
   /**
    * Indicate whether any uncommitted modification is being tracked for the
    * database table associated with the specified entity.  A modification is
    * the insertion, deletion, or update of a record.  Insertions and updates
    * considered the same for purposes of this method.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * 
    * @return  Bitfield indicating dirty status of entity, using constants
    *          from {@link DirtyInfo}.  Zero return indicates entity is not
    *          dirty.  Only <code>CHANGE</code> and <code>DELETE</code> are
    *          detected, not <code>INSERT</code>.
    */
   public int getEntityDirtyState(String entity)
   {
      int state = DirtyInfo.NONE;
      
      synchronized (changeLock)
      {
         if (changes.containsKey(entity))
         {
            state |= DirtyInfo.CHANGE;
         }
      }
      
      synchronized (deleteLock)
      {
         if (deletes.containsKey(entity))
         {
            state |= DirtyInfo.DELETE;
         }
      }
      
      return state;
   }
   
   /**
    * Indicate whether the record specified by the given entity and primary
    * key has been deleted within an uncommitted transaction in <i>any</i>
    * context.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the record to be checked.
    * 
    * @return  <code>true</code> if the specified record has been deleted,
    *          else <code>false</code>.
    */
   public boolean isDeleted(String entity, Serializable id)
   {
      synchronized (deleteLock)
      {
         Set<Serializable> deleted = deletes.get(entity);
         
         return (deleted != null && deleted.contains(id));
      }
   }
   
   /**
    * Mark a record as having been deleted in an uncommitted transaction, or
    * unmark a record which previously was so marked.
    * <p>
    * A record is unmarked when the adding context's current transaction is
    * committed or rolled back.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of deleted record.
    * @param   rollback
    *          If <code>false</code>, the target record is marked as deleted;
    *          if <code>true</code>, it is unmarked.
    * @param   lock
    *          <code>true</code> to acquire write locks to delete the record;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if there was an error deleting the record from the database.
    */
   public void delete(String entity, Serializable id, boolean rollback, boolean lock)
   throws PersistenceException
   {
      synchronized (deleteLock)
      {
         Set<Serializable> deleted = deletes.get(entity);
         
         if (rollback)
         {
            if (deleted != null)
            {
               deleted.remove(id);
            }
         }
         else
         {
            if (deleted == null)
            {
               deleted = new HashSet<Serializable>();
               deletes.put(entity.intern(), deleted);
            }
            deleted.add(id);
         }
      }
      
      if (!rollback)
      {
         boolean purgeFromDB = false;
         
         synchronized (changeLock)
         {
            if (changes != null)
            {
               Set<Serializable> changed = changes.get(entity);
               purgeFromDB = (changed != null && changed.remove(id));
            }
         }
         
         if (purgeFromDB)
         {
            synchronized (insertLock)
            {
               RecordIdentifier ident = new RecordIdentifier(entity, id);
               inserts.remove(ident);
               unvalidatedInserts.remove(ident);
            }
            
            StringBuilder buf = new StringBuilder("delete from ");
            buf.append(entity);
            buf.append(" where id = ");
            buf.append(id);
            String hql = buf.toString();
            Session session = null;
            Transaction tx = null;
            try
            {
               session = sessionFactory.openSession();
               tx = session.beginTransaction();
               Query query = session.createQuery(hql);
               query.executeUpdate();
               tx.commit();
            }
            catch (HibernateException exc)
            {
               if (tx != null)
               {
                  tx.rollback();
               }
               
               throw new PersistenceException("Error deleting dirty record", exc);
            }
            finally
            {
               if (session != null)
               {
                  try
                  {
                     session.close();
                  }
                  catch (HibernateException exc)
                  {
                     if (LOG.isLoggable(Level.SEVERE))
                     {
                        LOG.log(Level.SEVERE, "Error closing session", exc);
                     }
                  }
               }
            }
         }
         
         eventManager.enqueueDelete(entity, id);
      }
   }
   
   /**
    * Retrieve from the dirty database the record specified by the given
    * entity and primary key, if it exists.
    * <p>
    * This method temporarily places a share lock on all indexes containing
    * the primary key column for the target table.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the record to be retrieved.
    * 
    * @return  Target record if it exists in the dirty database, else
    *          <code>null</code>.
    * 
    * @throws  PersistenceException
    *          never.
    */
   public Persistable getDirtyDMO(String entity, Serializable id)
   throws PersistenceException
   {
      // Quick out if DMO is not being tracked.  Avoids overhead of index
      // locking.
      if (!isTracked(entity, id))
      {
         return null;
      }
      
      List<String> indexes = indexHelper.getIndexesForProperty(entity, PK_PROPERTY);
      Persistable copy = null;
      
      try
      {
         lockMultiple(indexes, LockType.SHARE);
         
         // Check again, in case status changed before we could get lock(s).
         if (!isTracked(entity, id))
         {
            return null;
         }
         
         RecordIdentifier ident = new RecordIdentifier(entity, id);
         
         // Try to retrieve the DMO as a newly inserted record first.  This is
         // faster than going to the database.
         synchronized (insertLock)
         {
            Persistable dmo = inserts.get(ident);
            if (dmo != null)
            {
               copy = (Persistable) dmo.deepCopy();
            }
         }
         
         // If record was not among those newly inserted, then it must be one
         // that previously existed in the primary database, and is now being
         // modified.  Look in the database.
         if (copy == null)
         {
            Session session = null;
            Transaction tx = null;
            
            try
            {
               session = sessionFactory.openSession();
               tx = session.beginTransaction();
               Persistable dmo = (Persistable) session.get(entity, id);
               copy = (Persistable) dmo.deepCopy();
               tx.commit();
            }
            catch (HibernateException exc)
            {
               if (tx != null)
               {
                  tx.rollback();
               }
               
               DBUtils.handleException(dirtyDatabase, exc);
               
               throw new PersistenceException(exc);
            }
            finally
            {
               if (session != null)
               {
                  try
                  {
                     session.close();
                  }
                  catch (Exception exc)
                  {
                     DBUtils.handleException(dirtyDatabase, exc);
                  }
               }
            }
         }
      }
      finally
      {
         lockMultiple(indexes, LockType.NONE);
      }
      
      return copy;
   }
   
   /**
    * Given certain search criteria and possibly the primary key of a
    * potential DMO match in the primary database, report any significant
    * information currently being tracked for uncommitted transactions, which
    * might override data (or the lack thereof) found in the primary database.
    * <p>
    * If a {@link DirtyInfo} object is returned by this method, it will
    * contain overriding information, such as a dirty DMO which matched the
    * search criteria, whether that record was newly inserted, or the fact
    * that the candidate record found in the primary search has been deleted
    * or modified in an uncommitted transaction.  A <code>null</code> return
    * indicates no overriding information was available.
    * <p>
    * This method acquires a share lock on <code>index</code>.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   candidateID
    *          Primary key ID of DMO, if any, found in the primary database,
    *          which matches the current query criteria.
    * @param   index
    *          Name of index which governs the ordering of data in the current
    *          query.
    * @param   statements
    *          One or more HQL statements to be executed against the dirty
    *          database to find a record which matches the current query
    *          criteria.  The list is in order of most specific to least
    *          specific search criteria.
    * @param   params
    *          Query substitution parameters which match placeholders in the
    *          most specific HQL statement.
    * @param   types
    *          Hibernate data types of the query substitution parameters.
    * 
    * @return  An object which summarizes the findings, or <code>null</code>
    *          if no uncommitted change information was available for the
    *          given arguments.
    *          
    * @throws  PersistenceException
    *          if there was any database error.
    */
   public DirtyInfo getDirtyInfo(String entity,
                                 Serializable candidateID,
                                 String index,
                                 List<String> statements,
                                 Object[] params,
                                 Type[] types)
   throws PersistenceException
   {
      DirtyInfo info = null;
      Session session = null;
      Transaction tx = null;
      Persistable dmo = null;
      
      try
      {
         lockSingle(index, LockType.SHARE);
         
         session = sessionFactory.openSession();
         tx = session.beginTransaction();
         int fullCount = params.length;
         int callCount = statements.size();
         Object[] processedParams = (fullCount == 0 ? params : preprocessQueryParameters(params));
         for (int i = 0; dmo == null && i < callCount; i++)
         {
            int usedCount = fullCount - i;
            Object[] usedParams = new Object[usedCount];
            Type[] usedTypes = new Type[usedCount];
            System.arraycopy(processedParams, 0, usedParams, 0, usedCount);
            System.arraycopy(types, 0, usedTypes, 0, usedCount);
            String hql = statements.get(i);
            Query query = session.createQuery(hql);
            query.setMaxResults(1);
            
            for (int j = 0; j < usedParams.length; j++)
            {
               query.setParameter(Integer.toString(j), usedParams[j], usedTypes[j]);
            }
            
            Object result = query.uniqueResult();
            if (result != null)
            {
               if (result instanceof Persistable)
               {
                  dmo = (Persistable) result;
               }
               else
               {
                  Serializable id = (Serializable) result;
                  dmo = (Persistable) session.get(entity, id);
               }
            }
         }
         tx.commit();
         
         if (dmo != null)
         {
            info = new DirtyInfo();
            Persistable copy = null;
            Serializable id = dmo.getId();
            RecordIdentifier ident = new RecordIdentifier(entity, id);
            
            // If the record is newly inserted, retrieve the copy made at the
            // time the record was added to the dirty database;  this is the
            // version of data that other contexts should be able to "see".
            synchronized (insertLock)
            {
               copy = inserts.get(ident);
               if (copy != null)
               {
                  info.setInserted();
                  info.setDirtyDMOs(copy, dmo);
               }
            }
            
            // If the record is not newly inserted, make a copy of the current
            // record;  callers may have to retrieve the appropriate version
            // of the data from the primary database, as this copy may contain
            // newer information than those contexts should be able to "see".
            if (copy == null)
            {
               copy = (Persistable) dmo.deepCopy();
               info.setDirtyDMOs(copy, copy);
            }
         }
         
         if (candidateID != null)
         {
            synchronized (changeLock)
            {
               Set<Serializable> changed = changes.get(entity);
               if (changed != null && changed.contains(candidateID))
               {
                  if (info == null)
                  {
                     info = new DirtyInfo();
                  }
                  info.setModified();
               }
            }
         }
      }
      catch (HibernateException exc)
      {
         if (tx != null)
         {
            tx.rollback();
         }
         
         DBUtils.handleException(dirtyDatabase, exc);
         
         throw new PersistenceException(exc);
      }
      finally
      {
         lockSingle(index, LockType.NONE);
         if (session != null)
         {
            try
            {
               session.close();
            }
            catch (Exception exc)
            {
               DBUtils.handleException(dirtyDatabase, exc);
            }
         }
      }
      
      return info;
   }
   
   /**
    * Execute an HQL query and return the results as a list.  If the query
    * returns multiple results per row, each element in the list will be an
    * array of <code>Object</code>s.
    * <p>
    * No locking is attempted.
    *
    * @param   hql
    *          HQL query statement passed to Hibernate.  This method makes no
    *          assumptions as to the types of object(s) returned.
    * @param   params
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    * @param   maxResults
    *          The maximum number of elements to be returned in the list.
    *          If this value is non-positive, no upper limit is applied.
    * @param   startOffset
    *          The 0-based offset of the first record to retrieve.  If this
    *          value is non-positive, an offset of 0 is used by default.
    * @param   readOnly
    *          <code>true</code> to execute the query in read-only mode, else
    *          <code>false</code>.
    *
    * @return  A list of results, or <code>null</code> if no result was found.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query.
    */
   public List list(String hql,
                    Object[] params,
                    Type[] types,
                    int maxResults,
                    int startOffset,
                    boolean readOnly)
   throws PersistenceException
   {
      Session session = null;
      Transaction tx = null;
      
      try
      {
         session = sessionFactory.openSession();
         tx = session.beginTransaction();
         Query query = session.createQuery(hql);
         Object[] processedParams = (params.length == 0
                                     ? params
                                     : preprocessQueryParameters(params));
         
         for (int i = 0; i < processedParams.length; i++)
         {
            query.setParameter(Integer.toString(i), processedParams[i], types[i]);
         }
         
         if (maxResults > 0)
         {
            query.setMaxResults(maxResults);
         }
         if (startOffset > 0)
         {
            query.setFirstResult(startOffset);
         }
         query.setReadOnly(readOnly);
         List results = query.list();
         tx.commit();
         
         return (results == null || results.isEmpty() ? null : results);
      }
      catch (HibernateException exc)
      {
         if (tx != null)
         {
            tx.rollback();
         }
         
         DBUtils.handleException(dirtyDatabase, exc);
         
         throw new PersistenceException(exc);
      }
      finally
      {
         if (session != null)
         {
            try
            {
               session.close();
            }
            catch (Exception exc)
            {
               DBUtils.handleException(dirtyDatabase, exc);
            }
         }
      }
   }
   
   /**
    * Stop tracking uncommitted changes to the the subset of records
    * represented by <code>subChanges</code>, and remove from the master set
    * of uncommitted, inserted records the subset of records represented by
    * <code>subInserts</code>.
    * 
    * @param   subChanges
    *          Sets of primary keys representing those records to be removed
    *          from uncommitted change tracking, mapped by entity name.
    * @param   subInserts
    *          A set of identifiers of those records to be removed from the
    *          master set of uncommitted inserts.
    * @param   leakedEntities
    *          Names of DMO types which have had their uncommitted updates
    *          "leaked" to other sessions via {@link #updateSnapshots(String,
    *          Persistable[])}.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void cleanupChanges(Map<String, Set<Serializable>> subChanges,
                              Set<RecordIdentifier> subInserts,
                              Set<String> leakedEntities)
   throws PersistenceException
   {
      if (subChanges == null)
      {
         return;
      }
      
      try
      {
         Iterator<Map.Entry<String, Set<Serializable>>> iter = subChanges.entrySet().iterator();
         while (iter.hasNext())
         {
            Map.Entry<String, Set<Serializable>> entry = iter.next();
            String entity = entry.getKey();
            Set<Serializable> ids = entry.getValue();
            removeRecords(entity, ids);
            if (leakedEntities.contains(entity))
            {
               synchronized (insertLock)
               {
                  boolean wasRemoved = false;
                  for (Serializable id : ids)
                  {
                     RecordIdentifier ri = new RecordIdentifier(entity, id);
                     if (inserts.remove(ri) != null)
                     {
                        wasRemoved = true;
                     }
                     unvalidatedInserts.remove(ri);
                  }
                  
                  if (wasRemoved && inserts.isEmpty())
                  {
                     inserts = new HashMap<>();
                     unvalidatedInserts = new HashMap<>();
                  }
               }
            }
         }
      }
      finally
      {
         if (subInserts != null)
         {
            synchronized (insertLock)
            {
               if (inserts.keySet().removeAll(subInserts) && inserts.isEmpty())
               {
                  inserts = new HashMap<>();
                  unvalidatedInserts = new HashMap<>();
               }
            }
         }
      }
   }
   
   /**
    * Remove from the master set of uncommitted, deleted records the subset of
    * records represented by <code>subDeletes</code>.
    * 
    * @param   subDeletes
    *          Sets of primary keys representing those records to be removed
    *          from uncommitted delete tracking, mapped by entity name.
    */
   public void cleanupDeletes(Map<String, Set<Serializable>> subDeletes)
   {
      if (subDeletes == null)
      {
         return;
      }
      
      synchronized (deleteLock)
      {
         Iterator<Map.Entry<String, Set<Serializable>>> iter = subDeletes.entrySet().iterator();
         while (iter.hasNext())
         {
            Map.Entry<String, Set<Serializable>> entry = iter.next();
            String entity = entry.getKey();
            Set<Serializable> ids = entry.getValue();
            Set<Serializable> deleted = deletes.get(entity);
            deleted.removeAll(ids);
            
            if (deleted.isEmpty())
            {
               deletes.remove(entity);
            }
         }
         
         if (deletes.isEmpty())
         {
            deletes = new HashMap<String, Set<Serializable>>();
         }
      }
   }
   
   /**
    * Stop tracking uncommitted changes to the the specified set of records
    * associated with the given entity name.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   ids
    *          Set of primary key IDs of the records to be removed from
    *          tracking.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   private void removeRecords(String entity, Set<Serializable> ids)
   throws PersistenceException
   {
      List<String> indexes = indexHelper.getIndexesForEntity(entity, false);
      Session session = null;
      Transaction tx = null;
      
      try
      {
         // Acquire the necessary write locks.
         lockMultiple(indexes, LockType.EXCLUSIVE);
         
         session = sessionFactory.openSession();
         tx = session.beginTransaction();
         
         Iterator<Serializable> iter = ids.iterator();
         for (int i = 0; iter.hasNext(); i++)
         {
            Serializable id = iter.next();
            Persistable dmo = (Persistable) session.get(entity, id);
            if (dmo != null)
            {
               session.delete(dmo);
            }
            else if (LOG.isLoggable(Level.WARNING))
            {
               StringBuilder buf = new StringBuilder("Attempt to remove ");
               buf.append(entity);
               buf.append("#");
               buf.append(id);
               buf.append(" from dirty database ");
               buf.append(dirtyDatabase);
               buf.append(" failed;  record not found in session");
               
               LOG.log(Level.WARNING, buf.toString());
            }
         }
         
         tx.commit();
      }
      catch (HibernateException exc)
      {
         if (tx != null)
         {
            tx.rollback();
         }
         
         DBUtils.handleException(dirtyDatabase, exc);
         
         throw new PersistenceException(exc);
      }
      finally
      {
         synchronized (changeLock)
         {
            Set<Serializable> changed = changes.get(entity);
            if (changed != null)
            {
               changed.removeAll(ids);
               if (changed.isEmpty())
               {
                  changes.remove(entity);
               }
            }
            
            if (changes.isEmpty())
            {
               changes = new HashMap<>();
            }
         }
         
         // Release locks.
         lockMultiple(indexes, LockType.NONE);
         
         if (session != null)
         {
            try
            {
               session.close();
            }
            catch (Exception exc)
            {
               DBUtils.handleException(dirtyDatabase, exc);
            }
         }
      }
   }
   
   /**
    * Attempt to acquire a lock of the type specified on a single index, or to
    * release an existing lock.
    * <p>
    * If invoked with a <code>lockType</code> that is not <code>NONE</code>,
    * this method will block if a target lock is currently unavailable
    * (<code>NO_WAIT</code> lock type variants are treated as regular,
    * blocking lock types).
    * 
    * @param   name
    *          Name of the index to lock.
    * @param   lockType
    *          Type of lock to acquire, or <code>NONE</code> to release an
    *          existing lock.  <code>NO_WAIT</code> variants will be treated
    *          as regular, blocking variants.
    * 
    * @see     com.goldencode.p2j.persist.lock.LockManager
    */
   private void lockSingle(String name, LockType lockType)
   {
      try
      {
         RecordIdentifier ident = new RecordIdentifier(name, INDEX_ID);
         lockManager.lock(lockType.toWaitVariant(), ident, true);
      }
      catch (LockUnavailableException exc)
      {
         // ok to ignore;  never thrown in this situation
      }
   }
   
   /**
    * Attempts to acquire multiple locks of the type specified, on the indexes
    * represented by <code>names</code>, or to release existing locks on those
    * indexes.
    * <p>
    * If invoked with a <code>lockType</code> that is not <code>NONE</code>,
    * this method will block if a target lock is currently unavailable
    * (<code>NO_WAIT</code> lock type variants are treated as regular,
    * blocking lock types).
    * 
    * @param   names
    *          Names of the indexes to lock.  To prevent deadlocks, this list
    *          should be consistently ordered for calls made across different
    *          contexts.  An ascending alphabetical sort is recommended.
    * @param   lockType
    *          Type of lock to acquire, or <code>NONE</code> to release
    *          existing locks.  <code>NO_WAIT</code> variants will be treated
    *          as regular, blocking variants.
    */
   private void lockMultiple(List<String> names, LockType lockType)
   {
      try
      {
         lockType = lockType.toWaitVariant();
         Iterator<String> iter = names.iterator();
         while (iter.hasNext())
         {
            String name = iter.next();
            RecordIdentifier ident = new RecordIdentifier(name, INDEX_ID);
            lockManager.lock(lockType, ident, true);
         }
      }
      catch (LockUnavailableException exc)
      {
         // ok to ignore;  never thrown in this situation
      }
   }
   
   /**
    * Retrieve the names of all indexes in which each of the given DMO
    * properties participate and merge them into a single list sorted in
    * ascending alphabetical order, discarding duplicates.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   properties
    *          Array of property names for which index names are to be
    *          retrieved.
    * 
    * @return  Alphabetically ascending list of index names.
    * 
    * @throws  PersistenceException
    *          if an error occurs querying index metadata.
    */
   private List<String> indexesForProperties(String entity, String[] properties)
   throws PersistenceException
   {
      List<String> list = null;
      int len = properties.length;
      if (len == 1)
      {
         // Only a single list.
         list = indexHelper.getIndexesForProperty(entity, properties[0]);
      }
      else
      {
         // Multiple lists which must be merged.  Duplicates are dropped and
         // the list is sorted alphabetically to prevent deadlocks.
         SortedSet<String> set = new TreeSet<String>();
         for (String property : properties)
         {
            list = indexHelper.getIndexesForProperty(entity, property);
            set.addAll(list);
         }
         list = new ArrayList<String>(set);
      }
      
      return list;
   }
   
   /**
    * Begin tracking changes which occur in the calling context's current
    * transaction for the given DMO.  This entails adding the DMO (a copy of
    * the original) into the dirty database and storing its primary key ID in
    * the mapping of changed records.  If the record was newly inserted into
    * its primary database, that fact is recorded as well.
    * <p>
    * This method acquires exclusive locks on all the indexes in the
    * <code>indexes</code> parameter.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   copy
    *          Copy of the DMO being tracked.
    * @param   lockIndexes
    *          List of indexes to lock while recording the uncommitted changes.  If
    *          <code>null</code>, no locking is attempted.
    * @param   properties
    *          Array of names of those DMO properties which have been updated.
    * @param   insert
    *          <code>true</code> if the record was newly inserted into its
    *          primary database, <code>false</code> if the record previously
    *          existed in the primary database and is simply being updated
    *          now.
    * @param   early
    *          <code>true</code> to indicate the insertion is happening at record creation, and
    *          the record has not yet been validated; <code>false</code> to indicate the insertion
    *          is happening at another point, such as a rollback of a deletion, or that this
    *          method is being called again upon validation of a new record, and the now-validated
    *          record should no longer be tracked as unvalidated.
    * @param   uniqueIndexes
    *          If <code>early</code> is <code>true</code>, this represents the unique indexes
    *          which must be checked for this record to fully pass validation. If
    *          <code>early</code> is <code>false</code>, this represents the unique indexes which
    *          were validated for this record for this change set.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   private void trackChange(String entity,
                            Persistable copy,
                            List<String> lockIndexes,
                            String[] properties,
                            boolean insert,
                            boolean early,
                            Collection<String> uniqueIndexes)
   throws PersistenceException
   {
      Serializable id = copy.getId();
      Session session = null;
      Transaction tx = null;
      
      try
      {
         createTable(entity);
         
         if (lockIndexes != null)
         {
            lockMultiple(lockIndexes, LockType.EXCLUSIVE);
         }
         
         session = sessionFactory.openSession();
         tx = session.beginTransaction();
         session.save(entity, copy);
         tx.commit();
         
         synchronized (changeLock)
         {
            Set<Serializable> changed = changes.get(entity);
            if (changed == null)
            {
               changed = new HashSet<Serializable>();
               changes.put(entity.intern(), changed);
            }
            changed.add(id);
         }
         
         if (insert)
         {
            RecordIdentifier ident = new RecordIdentifier(entity, id);
            synchronized (insertLock)
            {
               inserts.put(ident, copy);
               if (early)
               {
                  unvalidatedInserts.put(ident, new HashSet<>(uniqueIndexes));
               }
               else if (uniqueIndexes != null)
               {
                  Set<String> unchecked = unvalidatedInserts.get(ident);
                  if (unchecked != null &&
                      unchecked.removeAll(uniqueIndexes) &&
                      unchecked.isEmpty())
                  {
                     unvalidatedInserts.remove(ident);
                  }
               }
            }
            
            eventManager.enqueueInsert(entity, copy);
         }
         else
         {
            eventManager.enqueueUpdate(entity, copy, properties);
         }
      }
      catch (HibernateException exc)
      {
         if (tx != null)
         {
            tx.rollback();
         }
         
         DBUtils.handleException(dirtyDatabase, exc);
         
         throw new PersistenceException(exc);
      }
      finally
      {
         if (lockIndexes != null)
         {
            lockMultiple(lockIndexes, LockType.NONE);
         }
         if (session != null)
         {
            try
            {
               session.close();
            }
            catch (Exception exc)
            {
               DBUtils.handleException(dirtyDatabase, exc);
            }
         }
      }
   }
   
   /**
    * Apply the given property updates to the given DMO using reflection.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   dmo
    *          DMO being updated.
    * @param   properties
    *          Array of names of those DMO properties which are being updated.
    * @param   values
    *          Updated DMO property values.  Each value in the array
    *          corresponds positionally with its matching property name in the
    *          <code>properties</code> array.
    *          
    * @throws  PersistenceException
    *          if there is any error applying updates.
    */
   private void applyUpdates(String entity,
                             Persistable dmo,
                             String[] properties,
                             BaseDataType[] values)
   throws PersistenceException
   {
      Map<String, Method> setters = PropertyHelper.settersByProperty(entity);
      try
      {
         int len = properties.length;
         for (int i = 0; i < len; i++)
         {
            String property = properties[i];
            BaseDataType value = values[i];
            Method method = setters.get(property);
            method.invoke(dmo, value);
         }
      }
      catch (Exception exc)
      {
         throw new PersistenceException("Error storing dirty update(s)", exc);
      }
   }
   
   /**
    * Preprocess an array of query parameters appropriate to the dirty
    * database's dialect.
    * 
    * @param   params
    *          Query parameters.  If none, this should be a 0-length array.
    * 
    * @return  Preprocessed query parameters.
    */
   private Object[] preprocessQueryParameters(Object[] params)
   {
      Object[] args = params;
      int length = params.length;
      if (length > 0)
      {
         args = new Object[length];
         for (int i = 0; i < length; i++)
         {
            args[i] = Persistence.preprocessQueryParameter(params[i], dialect);
         }
      }
      
      return args;
   }
   
   /**
    * Given the name of an entity representing a database table and DMO type,
    * this method will create the "dirty" table and its associated indexes,
    * within the associated "dirty" database.  If the table already exists,
    * this method does nothing.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with the table being created.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   private void createTable(String entity)
   throws PersistenceException
   {
      synchronized (createdTables)
      {
         if (!createdTables.add(entity))
         {
            // Already created;  nothing to do.
            return;
         }
         
         Class<?> dmoClass = DBUtils.dmoClassForEntity(entity);
         Class<?> dmoIface = DMOIndex.getDMOInterface(dmoClass);
         Session session = null;
         Connection conn = null;
         Statement stmt = null;
         Transaction tx = null;
         
         try
         {
            session = sessionFactory.openSession();
            tx = session.beginTransaction();
            conn = ((SessionImpl) session).connection();
            stmt = conn.createStatement();
            
            // Create table(s).
            Iterator<String> tableIter = TempTableHelper.sqlTempTableCreate(dirtyDatabase,
                                                                            dmoIface);
            while (tableIter.hasNext())
            {
               String sql = tableIter.next();
               stmt.execute(sql);
            }
            
            // Create indexes (if any).
            Iterator<String> indexIter = TempTableHelper.sqlTempTableIndexCreate(dirtyDatabase,
                                                                                 dmoIface);
            while (indexIter.hasNext())
            {
               String sql = indexIter.next();
               stmt.execute(sql);
            }
            
            // Commit the changes.
            tx.commit();
         }
         catch (Exception exc)
         {
            if (tx != null)
            {
               tx.rollback();
            }
            createdTables.remove(entity);
            
            String msg = "Unable to create table for entity '" +
                         entity +
                         "' in database '" +
                         dirtyDatabase +
                         "'";
            
            DBUtils.handleException(dirtyDatabase, exc);
            
            throw new PersistenceException(msg, exc);
         }
         finally
         {
            if (stmt != null)
            {
               try
               {
                  stmt.close();
               }
               catch (SQLException exc)
               {
                  DBUtils.handleException(dirtyDatabase, exc);
                  if (LOG.isLoggable(Level.WARNING))
                  {
                     LOG.log(Level.WARNING, "Error", exc);
                  }
               }
            }
            
            if (session != null)
            {
               try 
               {
                  session.close();
               }
               catch (Exception exc)
               {
                  DBUtils.handleException(dirtyDatabase, exc);
                  if (LOG.isLoggable(Level.WARNING))
                  {
                     LOG.log(Level.WARNING, "Error", exc);
                  }
               }
            }
         }
      }
   }
}
