/*
** Module   : NopDirtyShareContext.java
** Abstract : Do-nothing implementation of DirtyShareContext interface
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20080520   @38632 Created initial version. A do-nothing
**                           implementation of DirtyShareContext interface.
** 002 CA  20080716   @39118 Renamed method lockUniqueIndexes to 
**                           lockAllIndexes - this will lock all the indexes 
**                           for a certain entity. Modified the delete(...)
**                           and rollbackInsert(...) methods so the index
**                           locking can be disabled.
** 003 ECF 20080815   @39545 Added support for global change events. These are
**                           events that represent index-changing inserts,
**                           updates, or deletes, which are processed by
**                           sessions other than the session in which the
**                           event originated.
** 004 ECF 20090428   @42132 Added isLocalInsert(). Reports whether a record
**                           is being tracked as newly inserted by this
**                           context.
** 005 ECF 20090601   @42572 Added support for Progress isolation leak quirk.
**                           Added new methods.
** 006 ECF 20090610   @42644 Modified updateSnapshots() method. Now returns
**                           list of DMOs affected.
** 007 ECF 20090819   @43694 Added new update() variant required by change to
**                           DirtyShareContext.
** 008 ECF 20131028          Import change.
** 009 ECF 20140821          Changed insert API to allow insertion of unvalidated, newly created
**                           records.
*/

package com.goldencode.p2j.persist.dirty;

import java.io.*;
import java.lang.reflect.Method;
import java.util.*;
import org.hibernate.type.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.util.*;

/**
 * An implementation of the {@link DirtyShareContext} interface which does
 * nothing.  This implementation is intended be used for temporary tables
 * (which do not share dirty data, since they are private to a session), and
 * possibly for database implementations which do not need dirty data sharing
 * services. 
 */
final class NopDirtyShareContext
implements DirtyShareContext
{
   /**
    * Register with a global event clearinghouse an interest in insert,
    * delete, and update events which affect the given DMO entity.
    * 
    * @param   entity
    *          DMO entity for which the caller is interested in tracking
    *          change.
    * 
    * @return  A unique identifier for the registrant, assigned by the event
    *          clearinghouse, used later to collect events and deregister.
    * 
    * @see     #getGlobalEvents(long, long)
    * @see     #deregisterForGlobalEvents(long)
    */
   public Long registerForGlobalEvents(String entity)
   {
      return null;
   }
   
   /**
    * Deregister with a global event clearinghouse an interest in insert,
    * delete, and update events which affect the given array of indexes.
    * 
    * @param   registerID
    *          A unique identifier for the registrant being deregistered.
    * 
    * @see     #getGlobalEvents(long, long)
    * @see     #registerForGlobalEvents(String)
    */
   public void deregisterForGlobalEvents(long registerID)
   {
   }
   
   /**
    * Retrieve any new events that have been collected by a global event
    * clearinghouse, representing inserts, deletes, and updates made to
    * records which participate in the indices for which the caller previously
    * has registered interest.
    * 
    * @param   registerID
    *          A unique identifier for an object which has registered interest
    *          in index-changing events.
    * @param   lastID
    *          A unique identifier representing the last event retrieved by
    *          the caller.  If less than 0, it is assumed no events have yet
    *          been retrieved since the caller registered interest in these
    *          events.
    * 
    * @return  Array of applicable, new events, or <code>null</code> if no
    *          such events have been collected.
    * 
    * @throws  EventRegistrationException
    *          if <code>registerID</code> is invalid or has expired.
    */
   public GlobalChangeEvent[] getGlobalEvents(long registerID, long lastID)
   throws EventRegistrationException
   {
      return null;
   }
   
   /**
    * Lock or unlock all indexes associated with the given DMO entity.
    * <p>
    * This implementation does nothing.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   lockType
    *          Type of lock to acquire, or <code>NONE</code> to release
    *          existing locks.  <code>NO_WAIT</code> variants will be treated
    *          as regular, blocking variants.
    * 
    * @throws  PersistenceException
    *          if there is an error querying JDBC metadata while retrieving
    *          index information.
    */
   public void lockAllIndexes(String entity, LockType lockType)
   throws PersistenceException
   {
   }
   
   /**
    * Begin tracking a newly inserted record in the dirty database.
    * <p>
    * This record is removed when the adding context's current transaction is
    * committed or rolled back.
    * 
    * @param   buffer
    *          Record buffer responsible for the insert.
    * @param   dmo
    *          Record to be tracked.
    * @param   setter
    *          Optional setter method for a property which must be updated in
    *          the DMO copy before it is inserted into the dirty database.
    *          May be <code>null</code>.
    * @param   datum
    *          Updated value to apply to <code>property</code>.  May be
    *          <code>null</code>.
    * @param   uniqueIndexes
    *          If the insertion is happening at record creation, and the record has not yet been
    *          validated, this will represent the collection of unique indexes that will need to
    *          be validated before the record can safely be persisted. If <code>null</code>, this
    *          indicates the record has been fully validated and should no longer be tracked as
    *          unvalidated.
    * @param   lock
    *          <code>true</code> to acquire write locks to insert the record;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void insert(RecordBuffer buffer,
                      Persistable dmo,
                      Method setter,
                      BaseDataType datum,
                      Collection<String> uniqueIndexes,
                      boolean lock)
   throws PersistenceException
   {
   }
   
   /**
    * Indicate whether the specified record is being tracked as newly inserted
    * by this context.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the record to be checked.
    * 
    * @return  <code>false</code>.
    */
   public boolean isLocalInsert(String entity, Serializable id)
   {
      return false;
   }
   
   /**
    * Indicate whether the given record identifier represents an unvalidated, newly inserted
    * record, in a different context than that of the caller. Assumes appropriate indexes already
    * locked by the caller.
    * 
    * @param   ident
    *          Record identifier (using entity name and primary key).
    * 
    * @return  <code>false</code>.
    */
   public boolean isForeignUnvalidatedIndex(RecordIdentifier ident, String index)
   {
      return false;
   }
   
   /**
    * If the specified record is an unvalidated insert created by a buffer local to this context,
    * get the buffer which created it.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation class associated
    *          with a table being tracked for uncommitted changes.
    * @param   id
    *          Primary key of the record to be checked.
    * 
    * @return  Creating buffer, or <code>null</code> if the record is not a locally created,
    *          unvalidated insert.
    */
   public RecordBuffer getCreatingBuffer(String entity, Serializable id)
   {
      return null;
   }
   
   /**
    * Rollback the insert of a record previously introduced with the {@link
    * #insert(RecordBuffer, Persistable, Method, BaseDataType, boolean, boolean)}
    * method.
    * <p>
    * This implementation does nothing.
    * 
    * @param   ident
    *          Identifier for the target record, which encapsulates its entity
    *          name and primary key.
    * @param   lock
    *          <code>true</code> to acquire write locks;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void rollbackInsert(RecordIdentifier ident, boolean lock)
   throws PersistenceException
   {
   }
   
   /**
    * Track an update to one or more indexed, DMO properties, as indicated by
    * the <code>properties</code> and <code>values</code> arguments.
    * <p>
    * Any record added to the dirty database as a result of this method is
    * removed when the adding context's current transaction is committed or
    * rolled back.
    * <p>
    * This implementation does nothing.
    * 
    * @param   buffer
    *          Record buffer responsible for the update.
    * @param   properties
    *          Array of names of those DMO properties which were updated.
    *          These should represent columns which participate in one or more
    *          indexes.
    * @param   values
    *          Updated DMO property values.  Each value in the array
    *          corresponds positionally with its matching property name in the
    *          <code>properties</code> array.
    * @param   validatedIndexes
    *          Collection of unique indexes which have been checked successfully for this update
    *          to have passed validation.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void update(RecordBuffer buffer,
                      String[] properties,
                      BaseDataType[] values,
                      Collection<String> validatedIndexes)
   throws PersistenceException
   {
   }
   
   /**
    * Track an update to one or more indexed, DMO properties, as indicated by
    * the <code>properties</code> and <code>values</code> arguments.
    * <p>
    * It is assumed the record identified by <code>entity</code> and
    * <code>id</code> already is present in the dirty database due to a
    * previous update or insert event.
    * <p>
    * This implementation does nothing.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the target record.
    * @param   properties
    *          Array of names of those DMO properties which were updated.
    *          These should represent columns which participate in one or more
    *          indexes.
    * @param   values
    *          Updated DMO property values.  Each value in the array
    *          corresponds positionally with its matching property name in the
    *          <code>properties</code> array.
    * @param   lock
    *          <code>true</code> to acquire write locks to update the record;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void update(String entity,
                      Serializable id,
                      String[] properties,
                      BaseDataType[] values,
                      boolean lock)
   throws PersistenceException
   {
   }
   
   /**
    * Overwrite the state of the records in the dirty database which match the
    * IDs of all DMOs for the given entities, which have been modified by the
    * current context.
    * <p>
    * This is necessary to support a quirk/bug in Progress which compromises
    * transaction isolation by leaking uncommitted changes (even non-indexed
    * data) under certain conditions.
    * <p>
    * This implementation does nothing.
    * 
    * @param   entities
    *          List of entities, which are the names of the DMO implementation
    *          classes associated with tables being tracked for uncommitted
    *          changes.  Only those records which match these entities will
    *          have their current state leaked.
    * @param   persistence
    *          Persistence service object to be used to retrieved the latest
    *          version of the target records in the current context.
    * 
    * @return  List of all DMOs whose snapshots were updated with the dirty
    *          share manager.  The list may be empty, but will not be
    *          <code>null</code>.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public List<Persistable> updateSnapshots(List<String> entities, Persistence persistence)
   throws PersistenceException
   {
      return Collections.emptyList();
   }
   
   /**
    * Mark a record as having been deleted in an uncommitted transaction, or
    * unmark a record which previously was so marked.
    * <p>
    * If marking a record as deleted, that was earlier inserted by the current
    * context, rollback the insert, but do not mark the record as deleted.
    * <p>
    * A record is unmarked when the adding context's current transaction is
    * committed or rolled back.
    * <p>
    * This implementation does nothing.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of deleted record.
    * @param   rollback
    *          If <code>false</code>, the target record is marked as deleted;
    *          if <code>true</code>, it is unmarked.
    * @param   lock
    *          <code>true</code> to acquire write locks to delete the record;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if a database error occurs rolling back an earlier insert.
    */
   public void delete(String entity, Serializable id, boolean rollback, boolean lock)
   throws PersistenceException
   {
   }
   
   /**
    * Indicate whether there are currently any uncommitted changes (inserts,
    * updates, or deletes) in the table represented by the given DMO entity
    * name.
    * 
    * @param   entity
    *          Fully qualified name of the DMO entity to be checked.
    * 
    * @return  <code>true</code> if there any pending, uncommitted changes in
    *          the backing table, else <code>false</code>.  This
    *          implementation always returns <code>false</code>.
    */
   public boolean isEntityDirty(String entity)
   {
      return false;
   }
   
   /**
    * Indicate whether the record specified by the given entity and primary
    * key has been deleted within an uncommitted transaction within a context
    * <i>other than the current context</i>.
    * <p>
    * This implementation does nothing.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the record to be checked.
    * 
    * @return  <code>false</code> always.
    */
   public boolean isDirtyDelete(String entity, Serializable id)
   {
      return false;
   }
   
   /**
    * Retrieve from the dirty database the record specified by the given
    * entity and primary key, if it exists, and if it was not put there by the
    * current context.
    * <p>
    * This implementation does nothing.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the record to be retrieved.
    * 
    * @return  <code>null</code> always.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public Persistable getDirtyDMO(String entity, Serializable id)
   throws PersistenceException
   {
      return null;
   }
   
   /**
    * Execute an HQL query and return the results as a list.  If the query
    * returns multiple results per row, each element in the list will be an
    * array of <code>Object</code>s.
    * <p>
    * This implementation does nothing.
    *
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   hql
    *          HQL query statement passed to Hibernate.  This method makes no
    *          assumptions as to the types of object(s) returned.
    * @param   params
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    * @param   maxResults
    *          The maximum number of elements to be returned in the list.
    *          If this value is non-positive, no upper limit is applied.
    * @param   startOffset
    *          The 0-based offset of the first record to retrieve.  If this
    *          value is non-positive, an offset of 0 is used by default.
    * @param   readOnly
    *          <code>true</code> to execute the query in read-only mode, else
    *          <code>false</code>.
    *
    * @return  <code>null</code> always.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query.
    */
   public List list(String entity,
                    String hql,
                    Object[] params,
                    Type[] types,
                    int maxResults,
                    int startOffset,
                    boolean readOnly)
   throws PersistenceException
   {
      return null;
   }
   
   /**
    * Given certain search criteria and possibly a potential DMO match in the
    * primary database, report any significant information currently being
    * tracked for uncommitted transactions, which might override data (or the
    * lack thereof) found in the primary database.
    * <p>
    * If a {@link DirtyInfo} object is returned by this method, it will
    * contain overriding information, such as a dirty DMO which matched the
    * search criteria, whether that record was newly inserted, or the fact
    * that the candidate record found in the primary search has been deleted
    * or modified in an uncommitted transaction.  A <code>null</code> return
    * indicates no overriding information was available.
    * <p>
    * This implementation does nothing.
    * 
    * @param   buffer
    *          Record buffer associated with the current query.
    * @param   index
    *          Name of index which governs the ordering of data in the current query.
    * @param   queryType
    *          Type of query.
    * @param   bundle
    *          An object containing one or more HQL statements to be executed
    *          against the dirty database to find a record which matches the
    *          current query criteria.  The list is in order of most specific
    *          to least specific search criteria.
    * @param   params
    *          Query substitution parameters which match placeholders in the
    *          most specific HQL statement.
    * @param   types
    *          Hibernate data types of the query substitution parameters.
    * @param   found
    *          DMO, if any, found in the primary database, which matches the
    *          current query criteria.
    * @return  <code>null</code> always.
    *          
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public DirtyInfo getDirtyInfo(RecordBuffer buffer,
                                 String index,
                                 int queryType,
                                 HQLBundle bundle,
                                 Object[] params,
                                 Type[] types,
                                 Persistable found)
   throws PersistenceException
   {
      return null;
   }
   
   /**
    * Clean up all information contributed to the dirty share manager by this
    * context for the current transaction.  This method is called when a full
    * database transaction has been committed or rolled back against the
    * primary database.  It is also invoked when the client context ends.  It
    * removes all records it added to the dirty database for new inserts,
    * updates, and deletes.
    * <p>
    * This implementation does nothing.
    */
   public void cleanup()
   {
   }
}
